use std::{borrow::Cow, time::Duration};

use futures_util::{SinkExt, StreamExt};
use reqwest::StatusCode;
use tokio::{
    sync::broadcast::{Receiver, Sender},
    time::sleep,
};
use tokio_tungstenite::{
    connect_async,
    tungstenite::{
        self,
        protocol::{frame::coding::CloseCode, CloseFrame},
        ClientRequestBuilder,
    },
};

use crate::{state_machine::Event, BASE_URL};

pub async fn start(
    event_sender: Sender<Event>,
    mut event_receiver: Receiver<Event>,
) -> Result<(), anyhow::Error> {
    let mut writer = None;

    let uri: tungstenite::http::Uri = format!("{}/ws", BASE_URL.replace("https", "wss")).parse()?;

    loop {
        if let Ok(event) = &event_receiver.try_recv() {
            match event {
                Event::TokenReceived { token } => {
                    let builder = ClientRequestBuilder::new(uri.clone())
                        .with_header("Authorization", format!("Bearer {}", token));

                    let (socket, response) = connect_async(builder).await?;

                    if response.status() != StatusCode::SWITCHING_PROTOCOLS {
                        tracing::error!("{:#?}", response);
                        continue;
                    }

                    let (write, mut read) = socket.split();
                    writer = Some(write);

                    tokio::spawn(async move {
                        let message = match read.next().await {
                            Some(data) => match data {
                                Ok(message) => message,
                                Err(e) => {
                                    tracing::error!("{:?}", e);
                                    return;
                                }
                            },
                            None => return,
                        };

                        let data = message.to_text();
                        tracing::debug!("{:?}", data);
                    });

                    tracing::info!("Connected");
                    let _ = event_sender.send(Event::Connected);
                }
                Event::Logout => {
                    if let Some(mut write) = writer {
                        write
                            .send(tungstenite::Message::Close(Some(CloseFrame {
                                code: CloseCode::Normal,
                                reason: Cow::from("User Logout"),
                            })))
                            .await?;
                        writer = None;
                        tracing::debug!("Disconnected");
                        event_sender.send(Event::Disconnected)?;
                    }
                }
                Event::Quit => {
                    tracing::info!("Shutting down Client");
                    if let Some(mut write) = writer {
                        write
                            .send(tungstenite::Message::Close(Some(CloseFrame {
                                code: CloseCode::Normal,
                                reason: Cow::from("Application Shutdown"),
                            })))
                            .await?;
                        tracing::debug!("Disconnected");
                    }
                    break;
                }
                _ => {}
            }
        }
        sleep(Duration::from_millis(100)).await;
    }
    tracing::info!("Client Shutdown");
    Ok(())
}
