use std::time::Duration;

use tokio::{
    sync::broadcast::{Receiver, Sender},
    time::sleep,
};

use crate::{
    config::Config,
    models::{CodeChallengeMethod, CodeVerifier},
    oauth,
};

#[derive(Debug, Clone, PartialEq)]
pub enum State {
    Init,
    AppStart,
    Authenticate,
    Connect { token: String },
    WaitForSim,
    InSim,
}

#[derive(Debug, Clone, PartialEq)]
pub enum Event {
    Ready,
    StartAuthenticate,               // should not be string
    TokenReceived { token: String }, // AppStart and Authenticate can fire off TokenReceived to transition into Connect
    Connected, // Once connected to the socket, and properly authenticated, fire off Connected to transition to WaitForSim
    Disconnected, // If for whatever reason we're disconnected from the backend, we need to transition back to Connect
    SimConnected, // SimConnect is connected, we're in the world and ready to send data, transition to Running
    SimDisconnected, // SimConnect is disconnected, we've finished the flight and exited back to the menu, transition back to WaitForSim
    Logout,
    Quit,
}

impl State {
    pub async fn next(self, event: Event) -> State {
        match (self.clone(), event.clone()) {
            // (Current State, SomeEvent) => NextState
            (_, Event::Ready) => State::AppStart,
            (_, Event::Logout) => State::AppStart,
            (_, Event::StartAuthenticate) => Self::Authenticate, // Goto Authenticate

            (_, Event::TokenReceived { token }) => State::Connect { token },

            (_, Event::Connected) => State::WaitForSim, // Goto WaitForSim

            (_, Event::SimConnected) => todo!(), // Goto InSim

            (_, Event::Disconnected) => State::AppStart, // Goto Connect
            (_, Event::SimDisconnected) => State::WaitForSim, // Goto WaitForSim

            (_, Event::Quit) => todo!(), // All events can go into quit, to shutdown the application
        }
    }

    pub async fn run(&self, signal: Sender<Event>, config: Config) -> Result<(), anyhow::Error> {
        match self {
            State::Init => Ok(()),
            State::AppStart => {
                if let Some(token) = config.token() {
                    signal.send(Event::TokenReceived {
                        token: token.to_string(),
                    })?;
                } else {
                    signal.send(Event::StartAuthenticate)?;
                }
                Ok(())
            }
            State::Authenticate => {
                if config.open_browser() {
                    oauth::open_browser(config.clone())?;
                }

                Ok(())
            }
            State::Connect { .. } => Ok(()),
            State::WaitForSim => {
                tracing::info!("Waiting for sim!");
                Ok(())
            }
            State::InSim => Ok(()),
        }
    }
}

pub async fn start(
    config: Config,
    event_sender: Sender<Event>,
    mut event_receiver: Receiver<Event>,
) -> Result<(), anyhow::Error> {
    let mut state = State::Init;

    state.run(event_sender.clone(), config.clone()).await?;

    loop {
        if let Ok(event) = event_receiver.try_recv() {
            state = state.next(event.clone()).await;

            state.run(event_sender.clone(), config.clone()).await?;

            if event == Event::Quit {
                tracing::info!("Shutting down State Machine");
                break;
            }
        }
        sleep(Duration::from_millis(100)).await;
    }

    tracing::info!("State Machine Shutdown");
    Ok(())
}
